//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright (c) Contributors to the OpenEXR Project.
//

#ifndef NAMESPACEALIAS_H_
#define NAMESPACEALIAS_H_

#include "IexNamespace.h"
#include "ImathNamespace.h"
#include "ImfNamespace.h"

namespace IMF   = OPENEXR_IMF_NAMESPACE;
namespace IMATH = IMATH_NAMESPACE;
namespace IEX   = IEX_NAMESPACE;

#endif /* NAMESPACEALIAS_H_ */
