//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright (c) Contributors to the OpenEXR Project.
//

//-----------------------------------------------------------------------------
//
//	class EnvmapAttribute
//
//-----------------------------------------------------------------------------

#define COMPILING_IMF_ENVMAP_ATTRIBUTE

#include <ImfEnvmapAttribute.h>

#if defined(_MSC_VER)
// suppress warning about non-exported base classes
#    pragma warning(disable : 4251)
#    pragma warning(disable : 4275)
#endif

OPENEXR_IMF_INTERNAL_NAMESPACE_SOURCE_ENTER

using namespace OPENEXR_IMF_INTERNAL_NAMESPACE;

template <>
IMF_EXPORT const char*
EnvmapAttribute::staticTypeName ()
{
    return "envmap";
}

template <>
IMF_EXPORT void
EnvmapAttribute::writeValueTo (
    OPENEXR_IMF_INTERNAL_NAMESPACE::OStream& os, int version) const
{
    unsigned char tmp = _value;
    Xdr::write<StreamIO> (os, tmp);
}

template <>
IMF_EXPORT void
EnvmapAttribute::readValueFrom (
    OPENEXR_IMF_INTERNAL_NAMESPACE::IStream& is, int size, int version)
{
    unsigned char tmp;
    Xdr::read<StreamIO> (is, tmp);
    _value = Envmap (tmp);
}

template <>
IMF_EXPORT void
EnvmapAttribute::copyValueFrom (
    const OPENEXR_IMF_INTERNAL_NAMESPACE::Attribute& other)
{
    _value = cast (other).value ();
}

template class IMF_EXPORT_TEMPLATE_INSTANCE
    TypedAttribute<OPENEXR_IMF_INTERNAL_NAMESPACE::Envmap>;

OPENEXR_IMF_INTERNAL_NAMESPACE_SOURCE_EXIT
